SystemOrganization addCategory: #'Tweak-Core-Object'!

Object subclass: #CFileOutMaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Core-Object'!

!CFileOutMaker commentStamp: 'ar 3/27/2004 15:11' prior: 0!
I am just a little tool that allows filing out the parts of Tweak in a well-defined order so they can be loaded again.!

!CFileOutMaker class methodsFor: 'fileOut' stamp: 'ar 10/4/2004 20:49'!
fileOutTweak
	"self fileOutTweak"
	| tweakCompiler tweakPreload tweakCore tweakFull file |
	file := FileStream forceNewFileNamed: 'Tweak-Organization.st'.
	(SystemOrganization categories select:[:each| each beginsWith: 'Tweak-']) do:[:catName|
		file cr; nextChunkPut: 'SystemOrganization addCategory: ', catName storeString.
	].
	file close.
	tweakCompiler := (PackageInfo named: 'Tweak-Compiler') asChangeSet.
	tweakPreload := (PackageInfo named: 'Tweak-Preload') asChangeSet.
	tweakCore := (PackageInfo named: 'Tweak-Core') asChangeSet.
	tweakFull := (PackageInfo named: 'Tweak') asChangeSet.
	tweakFull forgetAllChangesFoundIn: tweakCompiler.
	tweakFull forgetAllChangesFoundIn: tweakPreload.
	tweakFull forgetAllChangesFoundIn: tweakCore.
	tweakCore fileOut.
	tweakPreload fileOut.
	tweakCompiler fileOut.
	tweakFull fileOut.! !

Object subclass: #CNamespace
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Core-Object'!

CNamespace class
	instanceVariableNames: 'names uniqueInstance'!

!CNamespace commentStamp: 'ar 3/27/2004 17:12' prior: 0!
WARNING: THIS IS UNFINISHED AS OF YET!! But here's how it's supposed to work anyways:

Essentially, we don't want to screw with all of the existing tools in Squeak, so we just define that all classes are in the global namespace "Smalltalk" but that the names of the classes are really fully qualified ones. So, for example, the instance variable name 'name' of class Bar in namespace Foo will contain #'Foo::Bar' and we merely fix the lookup so that the namespace contains an entry under #Bar which is found by the compiler. This allows us to leave all of tools alone working with the name of #Foo::Bar if they aren't namespace-aware and have the tools which *are* namespace aware to use #localName or #fullName as they desire.

This scheme has various advantages: a) You can reflect about all classes in the system by enumerating through Smalltalk (useful for system-wide tools). b) When we print an object (or inspect a class or similar) we will see it's full name which is good as it avoids confusion. c) If we want to "import" some class it will know where it was defined so we can track that dependency accordingly. Etc. This seems *hugely* advantageous to approaches which try to keep the "local name" of the class and do all sorts of guessing about "who this guy is" (say, Environment>>scopeFor:from:envtAndPathIfFound:) - after all if a class doesn't know where it is defined, who the hell *would* know?

So the basic approach here is: Classes know who they are and where they are defined, and we can ask them for a local name which will be a "shorthand notation" in their namespace.

The stuff that really doesn't work right now is to fix the tools to the point that they can deal with namespaces AT ALL. There are so many freaking places in the system that are hardwired it's almost impossible to fix them all at once. For example, there's an interesting issue with respect to defining classes - those class definitions *have* to be executed by the environment that defines the class but unless we want to change the class definition message (which would screw up even more tools) you can't really guess which name is being used. Argh... anyways if you want to see what I'm talking about try to define something like:

Squeak::Object subclass: #Object
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Kernel'

inside the Tweak namespace. You'll see what I'm talking about...
!

CNamespace class
	instanceVariableNames: 'names uniqueInstance'!

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 14:13'!
at: aString
	^self names at: aString asSymbol! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 14:13'!
at: aString ifAbsent: aBlock
	^self names at: aString asSymbol ifAbsent: aBlock! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 14:13'!
at: aString ifPresent: aBlock
	^self names at: aString asSymbol ifPresent: aBlock! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 14:13'!
at: aString put: aValue
	^self names at: aString asSymbol put: aValue! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:58'!
bindingOf: varName
	| aSymbol binding |
	aSymbol := varName asSymbol.
	binding := self names bindingOf: aSymbol.
	binding ifNotNil:[^binding].
	"Pass to parent namespace"
	self == CNamespace
		ifFalse:[^superclass bindingOf: aSymbol].
	"For the root, we only resolve Squeak to alias Smalltalk for the time being"
	aSymbol == #Squeak ifTrue:[^Smalltalk associationAt: #Smalltalk].
	^nil! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:27'!
bindingsDo: aBlock
	^self names bindingsDo: aBlock! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:28'!
classBindingOf: varName
	"For initialization messages grant the regular scope"
	^super bindingOf: varName! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:28'!
classesDo: aBlock
	self bindingsDo:[:assoc| assoc value isBehavior ifTrue:[aBlock value: assoc value]].! !

!CNamespace class methodsFor: 'compiling' stamp: 'ar 3/27/2004 04:16'!
defaultCompilerClass
	"Answer the default compiler class for this name space"
	^Compiler! !

!CNamespace class methodsFor: 'hacks' stamp: 'ar 3/27/2004 06:02'!
defineClass: aClass
	"Define the given class in this environment. This is a hack 
	allowing Smalltalk to see classes defined in name spaces."
	| stName stBinding myName myBinding newName newBinding cat oldEnv |
	stName := aClass name.
	stBinding := Smalltalk bindingOf: stName.
	"Figure out the local name"
	myName := (stName findTokens:':') last.
	myBinding := self bindingOf: myName.
	(myBinding == nil or:[myBinding value == aClass]) ifFalse:[
		"Something screwed up"
		self notify: 'WARNING: ', myName, ' already exists.
Proceed will overwrite it.'.
	].
	newName := (self fullNameFor: myName) asSymbol.
	newBinding := Smalltalk bindingOf: newName.
	(newBinding == nil or:[newBinding value == aClass]) ifFalse:[
		"Something screwed up"
		self notify: 'WARNING: Smalltalk::',newName, ' already exists.
Proceed will overwrite it.'.
	].

	"Finally, really do it"
	cat := aClass category.
	oldEnv := aClass environment.
	aClass environment: self.
	self at: myName put: aClass.
	oldEnv organization removeElement: stName.
	self organization classify: myName under: cat.
	aClass setName: newName.

	stBinding ifNil:[
		Smalltalk at: newName put: aClass.
	] ifNotNil:[
		Smalltalk removeKey: stBinding key.
		stBinding key: newName.
		Smalltalk add: stBinding.
	].! !

!CNamespace class methodsFor: 'compiling' stamp: 'ar 3/27/2004 04:23'!
evaluate: aString notifying: aController logged: logFlag
	^self 
		defaultCompilerClass
		evaluate: aString 
		notifying: aController 
		logged: logFlag
		in: self! !

!CNamespace class methodsFor: 'accessing' stamp: 'ar 3/27/2004 04:46'!
fullName
	"Answer the fully qualified path for this namespace"
	| superName |
	self == CNamespace ifTrue:[^''].
	superName := superclass fullName.
	superName isEmpty ifTrue:[^name].
	^superName,'::', name
! !

!CNamespace class methodsFor: 'accessing' stamp: 'ar 3/27/2004 04:47'!
fullNameFor: localName
	"Answer the fully qualified name for the given local name"
	^(self fullName,'::', localName)
! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 14:13'!
includesKey: aName
	"does this pool include aName"
	^(self bindingOf: aName asSymbol) notNil! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:28'!
keyAtIdentityValue: obj ifAbsent: aBlock
	^self names keyAtIdentityValue: obj ifAbsent: aBlock! !

!CNamespace class methodsFor: 'changes' stamp: 'ar 3/27/2004 01:27'!
logChange: change
	^Smalltalk logChange: change! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:28'!
names
	^names ifNil:[names := IdentityDictionary new].! !

!CNamespace class methodsFor: 'accessing' stamp: 'ar 3/27/2004 01:26'!
obsolete
	self classesDo:[:aClass| aClass removeFromSystem: false].
	^super obsolete.! !

!CNamespace class methodsFor: 'accessing' stamp: 'ar 3/27/2004 01:26'!
organization
	^organization ifNil:[organization := SystemOrganizer defaultList: self names keys asArray].
! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 06:04'!
removeClassFromSystem: aClass logged: aBool
	"Delete the class, aClass, from the system, but log the removal neither to the current change set nor to the changes log"
	aBool ifTrue:[
		aClass wantsChangeSetLogging ifTrue:
			[ChangeSet current noteRemovalOf: aClass].
		aClass acceptsLoggingOfCompilation ifTrue:
			[Smalltalk logChange:  self name, ' removeClassNamed: #', aClass localName].
	].
	self organization removeElement: aClass localName.
	self names removeKey: aClass localName ifAbsent: [].
	Smalltalk removeClassFromSystem: aClass logged: false.! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 01:28'!
scopeFor: varName from: lower envtAndPathIfFound: envtAndPathBlock
	| binding path cls list |
	binding := self names bindingOf: varName.
	binding ifNil:[^nil].
	path := String streamContents:[:s|
		cls := self.
		list := #().
		[cls == CNamespace] whileFalse:[
			list := list copyWith: cls name.
			cls := cls superclass].
		list reversed do:[:scope|  s nextPutAll: scope; nextPutAll: '::'].
	].				
	^envtAndPathBlock
		value: self
		value: path.! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 13:58'!
template: aSystemCategoryName 
	"Answer an expression that can be edited and evaluated in order to define a new class."
	^self templateForSubclassOf: 'Squeak::Object' category: aSystemCategoryName! !

!CNamespace class methodsFor: 'lookup' stamp: 'ar 3/27/2004 13:58'!
templateForSubclassOf: priorClassName category: systemCategoryName 
	"Answer an expression that can be edited and evaluated in order to define a new class, given that the class previously looked at was as given"

	^priorClassName asString, ' subclass: #NameOfSubclass
	instanceVariableNames: ''''
	classVariableNames: ''''
	poolDictionaries: ''''
	category: ''' , systemCategoryName asString,''''! !

!CNamespace class methodsFor: 'accessing' stamp: 'ar 3/27/2004 04:22'!
uniqueInstance
	"Answer the unique instance of this name space"
	^uniqueInstance ifNil:[uniqueInstance := self new]! !

!Object methodsFor: '*Tweak-Core-Object' stamp: 'ar 5/7/2005 17:20'!
autoStartScripts
	"Start all the scripts that should be automatically initiated when the object is created"
	self myScripts size > 1 ifTrue: [self error: 'Scripts need to be empty here!!'].
	self class methodTriggers do:[:each| each installIn: self].! !

!Object methodsFor: '*Tweak-Core-Object' stamp: 'ar 5/7/2005 18:37'!
autoStartTriggers
	"Start all the scripts that should be automatically initiated when the object is created"
	self myScripts do:[:each| each autoStartTrigger].! !

!Object methodsFor: '*Tweak-Core-Object' stamp: 'ar 3/11/2004 19:43'!
definedFields
	^self class allFields, self instanceFields! !

!Object methodsFor: '*Tweak-Core-Object' stamp: 'eem 5/23/2008 11:03'!
get: fieldName
	| index |
	index := self class
				instVarIndexFor: fieldName
				ifAbsent:
					[^(self class bindingOf: fieldName) ifNotNil: [:var| var value]].
	^self instVarAt: index! !

!Object methodsFor: '*Tweak-Core-Object' stamp: ''!
getField: aField
	aField accessKey 
		ifNil:[^self perform: aField toGet]
		ifNotNil:[^self perform: aField toGet with: aField accessKey]! !

!Object methodsFor: '*Tweak-Core-Object' stamp: 'tk 2/7/2005 21:05'!
installNonTriggeredScripts: deepCopier
	"Pick out the non-triggered scripts and copy them to myself."

	deepCopier tweakScripts at: self ifPresent: [:myNonTriggered |
		myNonTriggered do: [:script |
			script myTrigger ifNil: ["created via #startScript:"
				script selector == #evaluate: ifTrue: [
					"create an identical script"
					self startScript: ((script arguments first) "a block"
										 veryDeepCopyWith: deepCopier)	 
						when: (script eventSpec veryDeepCopyWith: deepCopier)]].
							"it was already copied, just find the new one."
			script isPaused ifTrue: ["original is paused via #pauseScript:"
				self pauseScript: script selector]
			].
		].! !

!Object methodsFor: '*Tweak-Core-Object' stamp: ''!
instanceFields
	"Answer the fields defined only for this instance"
	self basicSize = 0 ifTrue:[^#()].
	^Array streamContents:[:s|
		1 to: self basicSize do:[:i|
			| fieldDef |
			fieldDef := CFieldDefinition named: i asString.
			fieldDef toGet: #basicAt:.
			fieldDef toSet: #basicAt:put:.
			fieldDef accessKey: i.
			s nextPut: fieldDef.
		].
	]! !

!Object methodsFor: '*Tweak-Core-Object' stamp: ''!
setField: aField to: aValue
	aField accessKey
		ifNil:[^self perform: aField toSet with: aValue]
		ifNotNil:[^self perform: aField toSet with: aField accessKey with: aValue]! !

Object subclass: #Tweak
	instanceVariableNames: ''
	classVariableNames: 'FullScreenMode'
	poolDictionaries: ''
	category: 'Tweak-Core-Object'!

!Tweak class methodsFor: 'compiling' stamp: 'ar 3/27/2004 04:16'!
defaultCompilerClass
	"Answer the default compiler class for this name space"
	^CCompiler! !

!Tweak class methodsFor: 'misc' stamp: 'ar 12/18/2004 22:04'!
fullScreenMode
	"Answer whether the system is in fullscreen mode"
	^FullScreenMode ifNil:[false].! !

!Tweak class methodsFor: 'misc' stamp: 'ar 12/18/2004 22:04'!
fullScreenMode: aBool
	"Enable or disable fullscreen mode"
	FullScreenMode := aBool.
	self primitiveSetFullScreen: aBool.! !

!Tweak class methodsFor: 'misc' stamp: 'ar 12/18/2004 19:44'!
primitiveSetFullScreen: aBool
	<primitive: 233>
	^self primitiveFailed! !

!Dictionary methodsFor: '*Tweak-Core-Object' stamp: 'tk 7/13/2004 16:56'!
atProperty: key put: newValue
	"Set the value at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer the previous value."

	| index assoc oldValue |
	index := self findElementOrNil: key.
	assoc := array at: index.
	assoc ifNil: [
		newValue == nil ifFalse: [
			"If no previous value, and new one is nil, don't record"
			self atNewIndex: index put: (Association key: key value: newValue)].
		^nil].
	oldValue := assoc value.
	assoc value: newValue.
	^oldValue! !

!Dictionary methodsFor: '*Tweak-Core-Object' stamp: 'ar 3/15/2004 21:02'!
removeBinding: key ifAbsent: aBlock
	"Remove the binding with the given key if it exists"
	| index assoc |
	index := self findElementOrNil: key.
	assoc := array at: index.
	assoc == nil ifTrue: [^ aBlock value].
	array at: index put: nil.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^assoc! !

Association subclass: #CComponentBinding
	instanceVariableNames: 'home'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Core-Object'!

!CComponentBinding commentStamp: 'ar 3/15/2004 21:21' prior: 0!
A component binding is an optimized representation of a two-way link between a component owner and the owned component.

Instance variables:
	key		<Symbol>	The name of the component (as known to its owner)
	value	<Object>		The component
	home	<Object>		The component owner.

The component binding is used by the owner to point to its component (retrieving the value) whereas it is stored in the component's componentBinding field and the component refers to its owner via the home of the binding.!

!CComponentBinding methodsFor: 'accessing' stamp: 'ar 1/22/2005 20:01'!
home
	^home! !

!CComponentBinding methodsFor: 'accessing' stamp: 'ar 3/15/2004 21:19'!
home: anObject
	home ifNotNil:[^self error: 'Cannot change home'].
	anObject == value ifTrue:[^self error:'Invalid ownership relation'].
	home := anObject.! !

!CComponentBinding methodsFor: 'testing' stamp: 'ar 3/15/2004 20:58'!
isComponentBinding
	^true! !

!CComponentBinding methodsFor: 'accessing' stamp: 'ar 3/15/2004 20:14'!
name
	^home name,'''s ', key! !

!CComponentBinding methodsFor: 'printing' stamp: 'ar 3/15/2004 20:14'!
printOn: aStream
	home printOn: aStream.
	aStream nextPutAll:'''s '; nextPutAll: key asString.
	aStream nextPutAll:' -> '; print: value.! !

!CComponentBinding methodsFor: 'tweaking' stamp: 'ar 1/29/2005 15:35'!
tweakedComponent: aComponent
	aComponent == value ifFalse:[^self error:'We *really* screwed up'].
	home tweaked: key from: value to: value.! !

!CComponentBinding methodsFor: 'accessing' stamp: 'ar 3/15/2004 21:19'!
value: aValue
	value == aValue ifTrue:[^value].
	aValue == home ifTrue:[^self error:'Invalid ownership relation'].
	aValue componentBinding: self.
	^value := aValue! !

CProtoObject subclass: #CObject
	instanceVariableNames: 'myProperties myScripts myEventMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Core-Object'!

!CObject commentStamp: 'tk 9/29/2004 10:38' prior: 0!
CObject defines the notions that any object in the Tweak universe should have.

Instance variables that begin with 'my' or that appear in red belong to the underlying mechanism.  DO NOT store into them or modify them.

myProperties		Dictionary of all virtual instance variables
myScripts			list of all scripts that currently have triggers to start them
myEventMap		Dictionary of (event name -> script name)
name				a String that names this object
value				the most important data this objects owns (whether it is checked for a CheckBox, etc.)
components			???!

CObject defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true" toGet="myProperties">myProperties</field>
	<field instanceVariable="true" toGet="myScripts" toSet="myScripts:">myScripts</field>
	<field instanceVariable="true" toGet="myEventMap" toSet="myEventMap:">myEventMap</field>
	<field toGet="name" toSet="name:" changeEvent="nameChanged">name</field>
	<field toGet="value" toSet="value:" changeEvent="valueChanged" serializer="writeValueOn:" deserializer="readValueFrom:">value</field>
	<field virtual="true" toGet="components">components</field>
</fields>'!

!CObject class methodsFor: 'fields' stamp: 'ar 3/11/2004 19:54'!
addField: fieldDef defaultValue: default
	"Add the field defined by fieldDef to the receiver."
	fieldDef isInstanceVariable ifTrue:[^self error: 'Not a field'].
	"hack around the problem of not having uniclasses yet"
	(fieldDef toGet == #propertyValueAt:) ifTrue:[
		fieldDef toGet: fieldDef name.
		fieldDef toSet: (fieldDef name,':') asSymbol.
		fieldDef accessKey: nil].
	self privateAddField: fieldDef.
	"compile accessors"
	self compileAccessorsFor: fieldDef defaultValue: default.! !

!CObject class methodsFor: 'fields' stamp: 'ar 11/18/2003 13:53'!
compileAccessorsFor: fieldDef defaultValue: default
	| code eventName |
	eventName := fieldDef name,'Changed'.
	code := String streamContents:[:s|
		s nextPutAll: fieldDef name.
		s crtab; nextPutAll: '"This method was automatically generated"'.
		s crtab; nextPutAll: '<autoGenerated: true>'.
		s crtab; nextPutAll: '<get: '; store: fieldDef name asSymbol; nextPutAll:'>'.
		s crtab; nextPutAll: '<signals: ', eventName,'>'.
		s crtab; nextPutAll: '<serialize: #writeProperty:on:>'.
		s crtab; nextPutAll: '<deserialize: #readProperty:from:>'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldDef name.
		default ifNotNil:[
			s nextPutAll: ' ifAbsent:['; store: default; nextPutAll:']'.
		].
	].
	self compile: code classified: 'field access'.
	code := String streamContents:[:s|
		s nextPutAll: fieldDef name; nextPutAll: ': aValue'.
		s crtab; nextPutAll: '"This method was automatically generated"'.
		s crtab; nextPutAll: '<autoGenerated: true>'.
		s crtab; nextPutAll: '<set: '; store: fieldDef name asSymbol; nextPutAll:'>'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldDef name.
		s nextPutAll: ' put: aValue'.
		s nextPutAll: ' with: #', eventName.
	].
	self compile: code classified: 'field access'.! !

!CObject class methodsFor: 'accessing' stamp: 'bf 4/27/2005 17:53'!
compileAllFrom: oldClass
	"HACK - provide interim methods for all selectors in oldClass that aren't here yet. Needed in order to make implicit self work properly with the compiler."
	oldClass selectorsDo:[:sel|
		(self includesSelector: sel) ifFalse:[
			self addSelectorSilently: sel withMethod: CompiledMethod toReturnSelf
		].
	].
	^super compileAllFrom: oldClass! !

!CObject class methodsFor: 'fields' stamp: 'ar 9/24/2003 14:17'!
compilePartAccessorFor: fieldDef
	| code eventName |
	eventName := fieldDef name,'Changed'.
	code := String streamContents:[:s|
		s nextPutAll: fieldDef name.
		s crtab; nextPutAll: '"This method was automatically generated"'.
		s crtab; nextPutAll: '<get: '; store: fieldDef name asSymbol; nextPutAll:'>'.
		s crtab; nextPutAll: '<signals: ', eventName,'>'.
		s crtab; nextPutAll: '<part: '; store: fieldDef name asSymbol; nextPutAll:'>'.
		s crtab; nextPutAll: '<serialize: #writeProperty:on:>'.
		s crtab; nextPutAll: '<deserialize: #readProperty:from:>'.
		s crtab; nextPutAll: '<category: #parts>'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldDef name.
	].
	self compile: code classified: 'parts access'.

	code := String streamContents:[:s|
		s nextPutAll: fieldDef name; nextPutAll: ': aValue'.
		s crtab; nextPutAll: '"This method was automatically generated"'.
		s crtab; nextPutAll: '<set: '; store: fieldDef name asSymbol; nextPutAll:'>'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldDef name.
		s nextPutAll: ' put: aValue'.
		s nextPutAll: ' with: #', eventName.
	].
	self compile: code classified: 'parts access'.! !

!CObject class methodsFor: 'fields' stamp: 'ar 3/12/2004 14:04'!
compilePropertyFor: fieldName defaultValue: default
	| code eventName |
	eventName := fieldName,'Changed'.
	code := String streamContents:[:s|
		s nextPutAll: fieldName.
		s crtab; nextPutAll: '"Answer the ', fieldName, ' of the receiver"'.
		s crtab; nextPutAll: '<bewareOf: #', eventName,'>'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldName.
		default ifNotNil:[
			s nextPutAll: ' ifAbsent:['; store: default; nextPutAll:']'.
		].
	].
	self compile: code classified: 'accessing'.
	code := String streamContents:[:s|
		s nextPutAll: fieldName; nextPutAll: ': aValue'.
		s crtab; nextPutAll: '"Modify the receiver''s ', fieldName, '"'.
		s crtab; nextPutAll: '^self propertyValueAt: #', fieldName.
		s nextPutAll: ' put: aValue'.
		s nextPutAll: ' with: #', eventName.
	].
	self compile: code classified: 'accessing'.! !

!CObject class methodsFor: 'copying' stamp: 'tk 1/18/2005 17:36'!
copyWeakly
	"If any fields (properties) have values should not be duplicated when I am duplicated, return the field names here.  Only mention CObjects that are not CPlayers, but are global shared data structures."

	^ #()! !

!CObject class methodsFor: 'instance creation' stamp: 'ar 9/25/2003 17:20'!
named: aString
	^(self new) name: aString; yourself! !

!CObject class methodsFor: 'instance creation' stamp: 'ar 5/7/2005 18:36'!
new
	^(self basicNew)
		autoStartScripts;
		initialize;			"user-level initialization"
		autoStartTriggers;
	yourself! !

!CObject class methodsFor: 'factory' stamp: 'ar 2/11/2003 15:26'!
next
	^self new! !

!CObject class methodsFor: 'instance creation' stamp: 'ar 2/7/2005 02:17'!
on: aValue
	^(self new) 	
		value: aValue; 
		onValueChanged;
	yourself! !

!CObject class methodsFor: 'fields' stamp: 'ar 5/26/2005 00:27'!
privateAddField: aFieldDefinition
	"Private. Add this field definition to the receiver. No error checking is done here."
	classFields ifNil:[classFields := #()].
	aFieldDefinition home: self.
	classFields := classFields copyWith: aFieldDefinition.
! !

!CObject class methodsFor: 'fields' stamp: 'ar 3/11/2004 19:56'!
privateRemoveField: aFieldDefinition
	"Private. Remove this field definition from the receiver. No error checking is done here."
	classFields ifNil:[classFields := #()].
	classFields := classFields copyWithout: aFieldDefinition.! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 21:12'!
addComponent: anObject named: componentName
	"Define the given object as being a component inside the receiver"
	| binding |
	"Create a new binding"
	binding := CComponentBinding new.
	binding home: self.
	binding key: componentName asSymbol.
	binding value: anObject.
	self addComponentBinding: binding.
	^anObject! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 21:03'!
addComponentBinding: aBinding
	"Add aBinding as a component binding to the receiver"
	| oldBinding |
	(aBinding isComponentBinding and:[aBinding home == self])
		ifFalse:[^self error: 'Invalid component binding'].
	myProperties ifNil:[myProperties := IdentityDictionary new].
	oldBinding := myProperties removeBinding: aBinding key ifAbsent:[nil].
	oldBinding ifNotNil:[oldBinding value: nil].
	myProperties add: aBinding.! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
addScriptor: aScriptor

	(self scriptors includes: aScriptor)
		ifFalse: [self propertyValueAt: #scriptors put: (self scriptors copyWith: aScriptor)]! !

!CObject methodsFor: 'initialize' stamp: 'ar 5/22/2005 10:55'!
becomeUniClass
	"Make the receiver an instance of a unique class"
	[self primitiveChangeClassTo: self class newUniClass basicNew] 
		on: CDeprecatedFeature do:[:ex| ex resume: true].
	"Need a GC to get rid of the instance we just created since #primitiveChangeClassTo: doesn't use the class but rather an instance. But I wonder if an IGC might be enough here?"
	Smalltalk garbageCollect.! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
compileScript: aString
	"Silently compile aString as a script for me as uniclass. Answer nil on failure, or the selector of the new script."

	| methodNode |
	methodNode := self class compilerClass new
		compile: aString in: self class notifying: nil
		ifFail: [^ nil].
	^self compileScriptNode: methodNode! !

!CObject methodsFor: 'scripting' stamp: 'bf 12/7/2005 14:47'!
compileScriptNode: aMethodNode
	"Become uniclass, add script, answer selector"

	| newMethod |
	self becomeUniClass.
	newMethod := aMethodNode generate: #(0 0 0 0).
	self class addSelectorSilently: aMethodNode selector withMethod: newMethod.
	self signal: #scriptModified with: aMethodNode selector.
	^ aMethodNode selector! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 20:09'!
componentBinding
	"Answer the receiver's component binding, e.g., the 'back link' to the component owner"
	<bewareOf: #componentBindingChanged>
	^self propertyValueAt: #componentBinding! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 20:09'!
componentBinding: aBinding
	"Change the component binding associated with the receiver."
	| myBinding |
	"Carefully release the previous binding first"
	(myBinding := self componentBinding) ifNotNil:[myBinding value: nil].
	^self propertyValueAt: #componentBinding put: aBinding with: #componentBindingChanged! !

!CObject methodsFor: 'components' stamp: 'tk 6/28/2004 12:33'!
componentHome
	"Answer the receiver's component owner (back link)"

	| cb |
	^ (cb := self propertyValueAt: #componentBinding) 
		ifNil: [nil]
		ifNotNil: [cb home]! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 20:11'!
componentNamed: aSymbol
	"Answer the component with the given name"
	^self componentNamed: aSymbol ifAbsent:[nil]! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 20:16'!
componentNamed: aSymbol ifAbsent: aBlock
	"Answer the component of the receiver with the given name"
	| binding |
	myProperties ifNil:[^aBlock value].
	binding := myProperties bindingOf: aSymbol.
	(binding == nil or:[binding isComponentBinding not]) ifTrue:[^aBlock value].
	^binding value! !

!CObject methodsFor: 'basic' stamp: 'ar 3/15/2004 21:25'!
components
	"Answer the components defined in the receiver."
	^self myComponents! !

!CObject methodsFor: 'copying' stamp: 'tk 1/13/2005 15:56'!
copy
	"Create a copy of the receiver.  The container pointer is not followed, but if my container is also being copied, then do use the new one for my copy.  "

	^self veryDeepCopy postCopy! !

!CObject methodsFor: 'basic' stamp: 'ar 3/1/2004 10:52'!
costume
	"Answer the receiver's costume"
	^nil! !

!CObject methodsFor: 'basic' stamp: 'ar 3/1/2004 10:52'!
costume: aCostume
	"Make the receiver wear the given costume"
	^nil! !

!CObject methodsFor: 'basic' stamp: 'ar 3/15/2004 21:12'!
define: componentName as: anObject
	"Define the given object as being a component inside the receiver"
	^self addComponent: anObject named: componentName! !

!CObject methodsFor: 'initialize' stamp: 'ar 3/21/2003 19:10'!
destroy
	"Destroy the receiver"
	self signal: #destroy with: self. "the last message you ever hear"
	self myEventMap: nil.
	self myScripts do:[:script| script terminateScript].
	self myScripts: nil.
! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
emptyScript
	"Replace the line above with the name of your script.  
	 Replace everything else with lines of script."! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
emptyScriptRef
	"Use this to include a new script in the list in a CScriptEdQuick"
	^ MethodReference new
		setStandardClass: CObject 
		methodSymbol: #emptyScript
! !

!CObject methodsFor: 'private' stamp: 'ar 5/22/2005 17:59'!
explorerContentsDo: aBlock
	self class allInstVarNames 
		keysAndValuesDo:[:index :iv| aBlock value: iv value: (self instVarAt: index)].
	myProperties ifNil:[^self].
	myProperties keysAndValuesDo: aBlock.! !

!CObject methodsFor: 'private' stamp: 'ar 1/15/2005 16:57'!
get: varName
	
	| aGetter var |
	(aGetter := varName) last == $: ifTrue: ["allow borderWidth:, and retrieve borderWidth"
		aGetter := (varName copyFrom: 1 to: (varName indexOf: $:)) asSymbol].
	^[self perform: aGetter] on: MessageNotUnderstood do:[:ex|
		ex return: ((var := self class bindingOf: varName) ifNotNil:[var value]).
	].! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
hasScripts
	"Answer true if this object has any scripts."

	self scriptsWithClassDo: [:sel :cls | ^true].
	^false! !

!CObject methodsFor: 'initialize' stamp: 'ar 9/25/2003 15:47'!
initialize
	"Initialize the receiver"! !

!CObject methodsFor: 'copying' stamp: 'tk 2/17/2005 12:14'!
installScriptsUsing: copier
	"Go through the scripts of my original object, and reconstruct them all here.  Ignore the class's list of scripts."

	| blk newSc got args script |
	copier tweakScripts at: self ifPresent: [:listOfScripts |
		myScripts == listOfScripts ifTrue: [self error: 'sharing scripts list'].
		myScripts ifNotNil: [myScripts size >= listOfScripts size ifTrue: [^ self]].	"second pass"
		listOfScripts do: [:script1 | "create an identical script"
			script := script1 clone.	"keep it steady while we run"
			got := false.
			script selector ifNil: [self error: 'still need to handle this case'].
			script myTrigger ifNotNil: [
				newSc := script myTrigger clone startUpIn: self.
					"maybe need to map any objects in the trigger?"
				got := true].
			got not & (script selector == #runTicking:at:) ifTrue: [got := true].
				"these should be started automatically"
			got not & (script selector == #evaluate:) ifTrue: [
				((blk := script arguments first) isKindOf: ContextPart) ifFalse: [
							"debug" self error: 'expecting block'].  
				newSc := self startScript: (blk veryDeepCopyWith: copier)	 
					when: (script eventSpec collect: [:aa | copier mapIfCan: aa]).
				got := true]. 	"were already copied, just find the new ones."
			got not & (script selector ~~ #evaluate:) ifTrue: [
				script extraArgs.	"computes them"
				args := script arguments allButLast: (script extraArgs min: script arguments size).
				newSc := self startScript: script selector 
					withArguments: (args collect: [:aa | copier mapIfCan: aa])
					when: (script eventSpec collect: [:aa | copier mapIfCan: aa]).
				got := true].

			script isPaused ifTrue: ["original is paused via #pauseScript:"
				self pauseScript: newSc selector].
			myScripts size > listOfScripts size ifTrue: [
					self error: 'more scripts than the original'].	"debug"
			]].
	! !

!CObject methodsFor: 'testing' stamp: 'tk 6/19/2004 17:02'!
isCObject

	^ true! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/1/2004 10:53'!
isTweaked
	"Answer whether the receiver has been tweaked by the user"
	self tweaks ifNotNil:[^true].
	^self costume isTweaked! !

!CObject methodsFor: 'testing' stamp: 'ar 3/1/2004 11:00'!
isUniClass
	"Answer whether I am a uniquely scripted object"
	^self class isUniClass! !

!CObject methodsFor: 'scripting' stamp: 'ar 1/22/2005 17:17'!
listOfScripts
	"Return an array of selectors for the scripts of this object."

	^Array streamContents: [:strm |
		self scriptsWithClassDo: [:sel :cls |
			strm nextPut: sel]].
! !

!CObject methodsFor: 'private' stamp: 'ar 10/18/2004 20:52'!
longPrintOn: aStream
	"Append to the argument, aStream, the names and values of all 
	of the receiver's instance variables."

	self class allFields do:[:field|
		(field isVirtualField or:[field toGet isNil]) ifFalse:[
			aStream nextPutAll: field name;
				nextPut: $:;
				space;
				tab;
			 	print: (self getField: field);
			 cr]].
	super longPrintOn: aStream.! !

!CObject methodsFor: 'scripting' stamp: 'bf 7/12/2005 18:35'!
lookup: aLookupSpec
	"Lookup object from aLookupSpec. See lookupSpecFor:"
	^self lookupScope lookup: aLookupSpec! !

!CObject methodsFor: 'scripting' stamp: 'bf 6/23/2005 20:08'!
lookupBehavior: aLookupSpec
	"Lookup a class from aLookupSpec. See lookupSpecFor:"
	self flag: #bf. "this should go away"
	^self class environment classNamed: aLookupSpec! !

!CObject methodsFor: 'scripting' stamp: 'bf 7/12/2005 18:38'!
lookupScope
	| env |
	env := Processor activeProject.
	env == self ifTrue: [self error: 'lookupScope must be overridden in objects handling lookup'].
	^ env! !

!CObject methodsFor: 'scripting' stamp: 'bf 7/12/2005 18:19'!
lookupSpecFor: anObject
	"Answer a literal that can later be used to lookup anObject. See lookup:"
	^self lookupScope lookupSpecFor: anObject! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 21:10'!
myComponents
	"Private. Answer the components of the receiver. This is severely screwed up below since what we *mean* to do when being asked for the components is to respond with a read-only mapped collection. Since there is no such thing (and I am too tired to write a proper one right now) we:
	a) Stupidly answer an empty id-dict if we have no properties at all. The reason we need to do this is that a client might ask for the 'keys' to get the names (Arrays don't answer their keys) and use it via #at:ifAbsent: (Array's screw up here).
	b) Use our own implementation of #collect: since Dictionary>>collect: is severely screwed up too (answering an OrderedCollection instead of what it *should* answer - another dictionary)."
	| result |
	myProperties ifNil:[^IdentityDictionary new].
	result := IdentityDictionary new.
	myProperties associationsDo:[:binding|
		binding isComponentBinding ifTrue:[result add: binding].
	].
	^result! !

!CObject methodsFor: 'private' stamp: 'ar 2/11/2003 15:19'!
myEventMap
	^myEventMap! !

!CObject methodsFor: 'private' stamp: 'ar 2/11/2003 15:19'!
myEventMap: anEventMap
	myEventMap := anEventMap.! !

!CObject methodsFor: 'private' stamp: 'tk 6/13/2004 13:10'!
myProperties
	^myProperties! !

!CObject methodsFor: 'private' stamp: 'ar 2/11/2003 15:19'!
myScripts
	^myScripts ifNil:[#()]! !

!CObject methodsFor: 'private' stamp: 'ar 2/11/2003 15:19'!
myScripts: scriptList
	myScripts := scriptList.! !

!CObject methodsFor: 'basic' stamp: 'bf 7/22/2004 12:40'!
name
	"Answer the name of the receiver"
	<bewareOf: #nameChanged>
	^(self propertyValueAt: #name) ifNil:[self class nameForDisplay]! !

!CObject methodsFor: 'basic' stamp: 'bf 3/30/2004 12:31'!
name: aString
	"Set the name of the receiver"
	^self propertyValueAt: #name put: aString with: #nameChanged.! !

!CObject methodsFor: 'initialize' stamp: 'ar 2/7/2005 02:17'!
onValueChanged
	"backstop"! !

!CObject methodsFor: 'basic' stamp: 'ar 3/1/2004 10:52'!
player
	"Answer the receiver's player"
	^nil! !

!CObject methodsFor: 'basic' stamp: 'ar 3/1/2004 10:52'!
player: aPlayer
	"Indicate the receiver's player"
	^nil! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/1/2004 10:56'!
postBuildWith: aMaker
	"Do any necessary post-construction activity"
	myScripts ifNotNil:[myScripts do:[:each| each resumeScript]].! !

!CObject methodsFor: 'copying' stamp: 'ar 3/1/2004 10:53'!
postCopy
	"Perform any necessary operations after copying completed"
	^self! !

!CObject methodsFor: 'private' stamp: 'ar 8/13/2003 00:43'!
printOn: aStream
	"| vv |
	vv := self value."
	aStream nextPutAll: self name.
	"vv == self ifFalse:[
		aStream nextPutAll: '<'; print: vv; nextPutAll:'>'.
	]."! !

!CObject methodsFor: 'properties' stamp: 'ar 3/1/2004 10:58'!
propertyValueAt: key
	"Answer the value of my property at key"
	myProperties ifNil:[^nil].
	^myProperties at: key ifAbsent:[nil]! !

!CObject methodsFor: 'properties' stamp: 'ar 3/1/2004 10:58'!
propertyValueAt: key ifAbsent: aBlock
	"Answer the value of my property at key"
	myProperties ifNil:[^aBlock value].
	^myProperties at: key ifAbsent: aBlock! !

!CObject methodsFor: 'properties' stamp: 'ar 3/1/2004 10:58'!
propertyValueAt: key ifAbsentPut: aBlock
	"Answer the value of my property at key"
	^self propertyValueAt: key ifAbsent:[
		self propertyValueAt: key put: aBlock value
	].! !

!CObject methodsFor: 'properties' stamp: 'ar 3/27/2004 21:10'!
propertyValueAt: key put: aValue
	"Store the value of my property at key"
	myProperties ifNil:[myProperties := IdentityDictionary new].
	myProperties at: key put: aValue.
	^aValue! !

!CObject methodsFor: 'properties' stamp: 'ar 9/8/2004 16:23'!
propertyValueAt: key put: newValue with: changeEvent
	"Store the value of my property at key"
	| oldValue |
	myProperties ifNil:[myProperties := IdentityDictionary new].
	oldValue := myProperties atProperty: key put: newValue.
	oldValue == newValue ifTrue:[^newValue].
	self signalChanged: changeEvent from: oldValue to: newValue.
	^newValue! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:02'!
readFieldsFrom: deserializer
	"Read my fields from the given deSerializer"
	self readFixedFieldsFrom: deserializer.
	self readPropertiesFrom: deserializer.! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:24'!
readFixedFieldsFrom: deserializer
	"Read all of the fixed fields from the deserializer. 
	Fixed fields in this context means all of the receiver's 
	instance variables which must be deserialized before
	the variable properties in the receiver."
	myScripts := deserializer readField: #scripts ifAbsent:[myScripts].! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:22'!
readPropertiesFrom: deserializer
	"Read my fields from the given deSerializer"
	| theClass fields theField |
	theClass := self class.
	[theClass == Object] whileFalse:[
		fields := theClass classFields.
		1 to: fields size do:[:i|
			theField := fields at: i.
			theField deserializer 
				ifNil:[self readProperty: theField name from: deserializer]
				ifNotNil:[self perform: theField deserializer with: deserializer].
		].
		theClass := theClass superclass
	].! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/27/2004 20:53'!
readProperty: propName from: deSerializer
	"Read the given property from deSerializer"
	| aValue |
	aValue := deSerializer readField: propName ifAbsent:[^nil].
	^self propertyValueAt: propName put: aValue! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/27/2004 20:44'!
readProperty: propName from: deSerializer with: changeEvent
	"Read the given property from deSerializer"
	| aValue |
	aValue := deSerializer readField: propName ifAbsent:[^nil].
	^self propertyValueAt: propName put: aValue" with: changeEvent"! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/1/2004 10:54'!
readTweakedFieldsFrom: deSerializer
	"Read all of the tweaked fields from the deSerializer"
	| field oldValue newValue |
	deSerializer fieldsDo:[:fieldName|
		field := self class fieldNamed: fieldName.
		field ifNotNil:[
			oldValue := self getField: field.
			newValue := deSerializer tweakField: fieldName value: oldValue.
			(newValue equals: oldValue) ifFalse:[
				self setField: field to: newValue.
				self tweaked: fieldName from: oldValue to: newValue.
			].
		].
	].! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/11/2004 03:14'!
readValueFrom: deSerializer
	"Read the my value from deSerializer"
	<deserialize: #value>
	self readProperty: #value from: deSerializer! !

!CObject methodsFor: 'components' stamp: 'ar 3/15/2004 20:25'!
removeComponentBinding: aBinding
	"Remove aBinding from my components."
	| oldBinding |
	(aBinding isComponentBinding and:[aBinding home == self])
		ifFalse:[^self error: 'Invalid component binding'].
	myProperties ifNil:[^self].
	oldBinding := myProperties removeKey: aBinding key ifAbsent:[nil].
	oldBinding ifNotNil:[oldBinding value: nil].! !

!CObject methodsFor: 'properties' stamp: 'ar 3/1/2004 10:59'!
removePropertyNamed: key
	"Remove the property with the given key"
	myProperties ifNil:[^nil].
	^myProperties removeKey: key ifAbsent:[nil]! !

!CObject methodsFor: 'scripting' stamp: 'bf 12/7/2005 14:47'!
removeScript: aSelector
	self assert: [self isUniClass].
	self removeScriptorFor: aSelector.
	self class removeSelectorSilently: aSelector.
	self signal: #scriptModified with: aSelector.
! !

!CObject methodsFor: 'scripting' stamp: 'bf 10/11/2004 18:15'!
removeScriptorFor: aSelector
	| scriptor scriptors |
	scriptor := self scriptorFor: aSelector.
	scriptor ifNotNil: [
		scriptors := self scriptors copyWithout: scriptor.
		scriptor destroy.
		scriptors isEmpty
			ifTrue: [self removePropertyNamed: #scriptors]
			ifFalse: [self propertyValueAt: #scriptors put: scriptors]].! !

!CObject methodsFor: 'scripting' stamp: 'bf 10/11/2004 18:07'!
scriptorFor: aSelector
	^self scriptors detect: [:each | aSelector = each selector]
		ifNone: [nil]! !

!CObject methodsFor: 'scripting' stamp: 'bf 10/7/2004 15:59'!
scriptors
	^self propertyValueAt: #scriptors ifAbsent: [#()] ! !

!CObject methodsFor: 'scripting' stamp: 'bf 3/4/2005 16:40'!
scriptsWithClassDo: aBlock
	"Evaluate aBlock with the selectors and defining classes of all scripts in this object.  Include superclasses that are uniclasses. Exclude DoIts and field accessors"

	self class withAllSuperclassesDo: [:cls | 
		cls isUniClass ifFalse: [^self].
		cls selectors do: [:sel |
			((#(DoIt DoItIn:) includes: sel)
				or: [cls classFields anySatisfy: [:f | sel == f toGet or: [sel == f toSet]]])
					ifFalse: [aBlock value: sel value: cls]]].
	self error: 'no non-uniclass in super chain'.! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/1/2004 10:57'!
serializeOn: aSerializer
	"Serialize the receiver on aSerializer"
	^aSerializer writeScripted: self.
! !

!CObject methodsFor: 'private' stamp: 'ar 3/11/2004 15:01'!
set: varName to: newValue
	^self perform: (varName,':') asSymbol with: newValue! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/15/2004 20:33'!
tweaked: propName from: oldValue to: newValue
	"Remember the tweak of the given property from its previous to its new value.
	Answer true if the receiver is newly changed or reverted back to unchanged."
	| owner |
	(super tweaked: propName from: oldValue to: newValue) ifFalse:[^false].
	"Note: we need to notify the player so the player can notify its home (if any).
	This feels wrong (don't ask me why)... but right now that seems to be the only
	way to make it work."
	(owner := self player) ifNotNil:[owner tweakedCostume: self].
	"Finally, notify my model if any"
	(owner := self componentBinding) ifNotNil:[owner tweakedComponent: self].
	^true! !

!CObject methodsFor: 'tweaking' stamp: 'ar 2/29/2004 00:41'!
tweakedCostume: aCostume
	"Broadcast tweaks happened to aCostume.
	This feels wrong ... see the senders"
	self costume == aCostume 
		ifTrue:[self tweaked: #costume from: aCostume to: aCostume].! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/1/2004 10:54'!
tweaks
	"Answer the tweaks associated with the receiver"
	^self propertyValueAt: #myTweaks! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/1/2004 10:54'!
tweaks: aDictionary
	"Set the tweaks associated with the receiver"
	^self propertyValueAt: #myTweaks put: aDictionary with: #tweaksChanged! !

!CObject methodsFor: 'user data' stamp: 'ar 3/24/2004 15:49'!
userData
	"Answer the receiver's userData properties. User data is intended for clients which want to associate certain idiosynchratic properties with objects in a specific context."
	^self propertyValueAt: #userData! !

!CObject methodsFor: 'user data' stamp: 'ar 3/24/2004 15:45'!
userData: aDictionary
	^self propertyValueAt: #userData put: aDictionary with: #userDataChanged! !

!CObject methodsFor: 'user data' stamp: 'bf 7/12/2005 18:53'!
userDataAt: aKey
	^(self userData ifNil:[^nil]) at: aKey ifAbsent: [nil]! !

!CObject methodsFor: 'user data' stamp: 'ar 3/24/2004 15:44'!
userDataAt: aKey ifAbsent: aBlock
	^(self userData ifNil:[^aBlock value]) at: aKey ifAbsent: aBlock! !

!CObject methodsFor: 'user data' stamp: 'ar 11/13/2004 18:02'!
userDataAt: aKey ifAbsentPut: aBlock
	^self userDataAt: aKey ifAbsent:[self userDataAt: aKey put: aBlock value]! !

!CObject methodsFor: 'user data' stamp: 'ar 3/24/2004 15:48'!
userDataAt: aKey put: aValue
	| data |
	data := self userData.
	data ifNil:[
		data := IdentityDictionary new.
		self userData: data.
	].
	aValue ifNil:[
		data removeKey: aKey ifAbsent:[].
		data isEmpty ifTrue:[^self userData: nil].
	].
	self signal: #userDataChanged.
	^data at: aKey put: aValue.! !

!CObject methodsFor: 'basic' stamp: 'ar 3/29/2004 00:38'!
value
	<bewareOf: #valueChanged>
	^self propertyValueAt: #value ifAbsent:[nil]! !

!CObject methodsFor: 'basic' stamp: 'tk 8/9/2004 18:10'!
value: newValue
	"Set the value field of this object.  Value is used to store the part of this object that matters most."

	^self propertyValueAt: #value put: newValue with: #valueChanged! !

!CObject methodsFor: 'copying' stamp: 'tk 1/20/2005 12:10'!
veryDeepCopyWith: copier
	"See veryDeepInner: and CPlayer veryDeepCopyWith:"

	copier newUniClasses: false. 	"Duplication never makes a new uniClass in Tweak."
			"set this every time just to make sure"
	^ super veryDeepCopyWith: copier! !

!CObject methodsFor: 'copying' stamp: 'bf 7/31/2005 21:11'!
veryDeepFixupWith: deepCopier 
	"Restart events and scripts.  See veryDeepInner:"

	super veryDeepFixupWith: deepCopier.

	self userDataAt: #lookupSpec put: nil.
	self installScriptsUsing: deepCopier.

"	self autoStartScriptsSafely.		rebuild triggered myScripts"
"	self installNonTriggeredScripts: deepCopier. "! !

!CObject methodsFor: 'copying' stamp: 'tk 2/16/2005 12:12'!
veryDeepInner: deepCopier 
	"Properties:  Just like any other tree of objects.  If the value is a CPlayer, copy it weakly.  It must be in elements to be duplicated.  If it is just a CObject, do duplicate it.  
	If I am a global data structure that should NOT be duplicated, mention the field name in its class in the message #copyWeakly.  See veryDeepFixupWith:"

	self myScripts size > 0 ifTrue: [
		deepCopier tweakScripts at: self put: self myScripts]. 	"save them all"

	super veryDeepInner: deepCopier.
	myProperties := self class copyWeakly size = 0 
		ifTrue: [myProperties veryDeepCopyWith: deepCopier]
		ifFalse: [myProperties veryDeepProperties: deepCopier].	"leave out forbidden fields"
	myScripts := nil.
	myEventMap := nil.
! !

!CObject methodsFor: 'copying' stamp: 'tk 1/24/2005 13:32'!
veryDeepProperties: deepCopier
	"Special case for when a class wants one of its fields to be shared and not duplicated."

	| old forbidden |
	(deepCopier newGuys includes: myProperties) ifTrue: [^ self].  	"already done"
	old := myProperties.
	myProperties := myProperties copy.	"has old values"  "does duplicate the associations"
	deepCopier references at: old put: myProperties.
	deepCopier newGuys add: myProperties.	"remember"

	"Skip some special fields"
	forbidden := self class copyWeakly.
	myProperties associationsDo: [:assoc |
			(forbidden includes: assoc key) ifFalse: [
				assoc value: (assoc value veryDeepCopyWith: deepCopier)].
			"do nothing if key is forbidden"
			deepCopier references at: (old associationAt: assoc key) put: assoc.
			deepCopier newGuys add: assoc].
! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:02'!
writeFieldsOn: serializer
	"Write my fields on the given serializer"
	self writeFixedFieldsOn: serializer.
	self writePropertiesOn: serializer.! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:43'!
writeFixedFieldsOn: serializer
	"Write all of the fixed fields on the serializer. 
	Fixed fields in this context means all of the receiver's 
	instance variables which must be serialized before
	the variable properties in the receiver."
	serializer writeField: #scripts value: myScripts.! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:01'!
writePropertiesOn: aSerializer
	"Write my property fields on the given serializer"
	| theClass fields theField |
	theClass := self class.
	[theClass == Object] whileFalse:[
		fields := theClass classFields.
		1 to: fields size do:[:i|
			theField := fields at: i.
			theField serializer 
				ifNil:[self writeProperty: theField name on: aSerializer]
				ifNotNil:[self perform: theField serializer with: aSerializer].
		].
		theClass := theClass superclass
	].! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/1/2004 10:58'!
writeProperty: propName on: serializer
	"Write the given property on the serializer"
	myProperties ifNil:[^self].
	serializer writeField: propName value: (myProperties at: propName ifAbsent:[^self]).! !

!CObject methodsFor: 'tweaking' stamp: 'ar 3/3/2004 16:40'!
writeTweakedFieldsOn: aSerializer
	"Write all of the tweaked fields on aSerializer"
	| changeList newValue |
	changeList := self tweaks ifNil:[^self].
	changeList removeKey: #costume ifAbsent:[].
	self costume isTweaked ifTrue:[
		aSerializer writeField: #costume value: self costume.
	].
	changeList keysAndValuesDo:[:fieldName :chgRec|
		newValue := self get: fieldName.
		aSerializer writeField: fieldName value: newValue.
	].! !

!CObject methodsFor: 'serializing' stamp: 'ar 3/11/2004 02:10'!
writeValueOn: aSerializer
	"Write my value on aSerializer"
	<serialize: #value>
	self writeProperty: #value on: aSerializer.! !
